subroutine ISAT_MP( nx, nf, ldxf, nv, xf, mode, nSR2, info_imp, istep )
!
! ------------ OVERVIEW 
! 
! The software ISAT_MP is an extension of the ISAT algorithm (Pope 1997) to the 
! multi-processor enviroment for efficient function evaluation. It tries to
! minimize the wall clock time spent in the evaluation of the computationally
! expensive function f(x) for an ensemble of x by using the x2f_mpi library
! and the serial ISAT algorithm. Given an ensemble of x, ISAT_MP returns the
! approximate value of f for each x which (with high probability) is within a
! specified tolerance of the exact value.
!
! ISAT_MP calls x2f_mpi, which contains three different distribution strategies
! and is responsible for distributing the ensemble of x among some or all of the 
! processors in the simulation, and then x2f_mpi calls ISAT, which performs the
! function evaluation. 
!
! In the multi-processor environment, before performing function evaluations,
! the ensemble of x may be distributed among some or all of the processors used
! in the simulation. Therefore some x originally located on one processor might
! be sent to the other processors and some x might stay locally. Then function
! evaluations are performed for all of the x resident on the processor at that
! time using the local ISAT table. After function evaluations, each f(x) is sent
! back to the processor from which the corresponding x originated.
!
! To evaluate f(x), the ISAT algorithm provides a sequence of methods, namely
! PR (primary retrieve), PR+SR (primary and secondary retrieve), and complete
! ISAT (including primary retrieve, secondary retrieve, grow, add and direct
! evaluation). The first two methods(PR and PR+SR) may or may not be successful
! in function evaluation, but they are cheap compared to the last one
! (complete ISAT) which always succeeds.
!
! In general, the computationally cheap methods are tried first to perform
! function evaluation. In the multi-processor environment, the retrieve
! methods can be tried on ISAT tables on different processors. By setting the
! values of atmps array in ISAT_MP, the user can specify the number of attempts
! for each of the methods, namely PR, PR+SR and complete ISAT. Setting atmps(i)=0
! corresponds to not using the method. Notice that more attempts on retrieval
! methods do not necessarily mean more efficiency.
!
! In x2f_mpi, three different distribution strategies are implemented, namely:
! purely local processing (PLP), uniform random distribution (URAN); and
! preferential distribution (PREF). Different distribution strategies are
! referenced by the index(e.g, 0 refers to PLP, 1 refers to URAN and 4 refers to
! PREF). In ISAT_MP, the user can specify the distribution strategies by setting
! the values in array dist_strat. By specifying dist_strat(i), the distribution
! strategy for the i-th method is set. 
!  
! When using ISAT_MP, the key parameters one needs to specify are the number of
! attempts for each method (atmps) and the distribution strategies for each method
! (dist_strat). In the current implementation, by setting mode=1, 2, 3 or 4, the
! user can choose one of the default settings for these two parameters. For
! advanced users, after setting mode=0, one can provide user-specified values of 
! atmps and dist_strat. 
!
! ------------ INPUT 
!
!  nx        - length of x vectors
!  nf        - length of f vectors
!  ldxf      - the leading dimension of the array xf ( ldxf >=max(nx,nf) )
!  nv        - number of vectors of x  
!  xf        - array which contains in xf(1:nx,1:nv) the vectors of x on input,
!            - and the vectors of f on output (see below)
! mode       - integer determing the action to be taken (see below)
! info_imp   - integer array dimensioned 10. The contents of info_imp(1:2) 
!              specify control parameters for using PREF strategy.  
! info_imp(1)= nolocal = 1, do not try ISAT table on local processor
!                      = 0, attempt to retrieve from ISAT table on local processor
! info_imp(2)= att_max, maximum number of attempts for retrieving method
! 
! info_imp(3:10) ---- reserved for future use
!
! ------------ OUTPUT
!
!  xf        - array which contains in xf(1:nf,1:nv) the vectors of f
!
! ------------ MODE
!
! There are some pre-set strategies which can be chosen by specifying a "mode" 
! (1<=mode<=4). For user-set strategy, set mode=0 and specify the distribution
! strategies (dist_strat) and number of attempts (atmps) for each method, also 
! modify k_pos to satisfy the requirements (see below).
!
!    mode    = 1 - PLP (purely local processing)
!                  with dist_strat(3)=(/0,0,0/) and 
!                       atmps(3)     =(/0,0,1/) specified
!            = 2 - URAN (uniform random distribution)
!                  with dist_strat(3)=(/0,0,1/) and
!                       atmps(3)     =(/0,0,1/) specified
!            = 3 - PREF(nproc)/URAN with nolocal=0
!                  with dist_strat(3)=(/0,4,1/) and 
!                       atmps(3)     =(/0,nproc,1/) specified
!                  nproc - the number of processor used in the simulation
!                - PREF(nproc-1)/URAN with nolocal=1
!                  with dist_strat(3)=(/0,4,1/) and 
!                       atmps(3)     =(/0,nproc-1,1/) specified
!                  nproc - the number of processor used in the simulation
!            = 4 - PREF(ntmp)/URAN with nolocal=0 and attempts limited to att_max
!                  with dist_strat(3)=(/0,4,1/) and 
!                       atmps(3)     =(/0,ntmp,1/) specified
!                  ntmp - determined by the number of processor(nproc), maximum
!                         number of limits(att_max) and nolocal
!                - PREF(ntmp)/URAN with nolocal=1 and attempts limited to att_max
!                  with dist_strat(3)=(/0,4,1/) and 
!                       atmps(3)     =(/0,ntmp,1/) specified
!                  ntmp - determined by the number of processor(nproc), maximum
!                         number of limits(att_max) and nolocal
!
! ------------ PARAMETERS
!
! ------------ Parameters must be specified by user only when mode=0
! dist_strat - distribution strategies used for each method, dist_strat(1:3)
! dist_strat(i) 
!            - indicates the distribution used for i-th method
!            = 0, PLP
!            = 1, URAN
!            = 4, PREF
! atmps      - number of attempts for each method, atmps(1:3)
! atmps(1)   - set the number of attempts for PR method
! atmps(2)   - set the number of attempts for PR+SR method
! atmps(3)   - set the number of attempts for complet ISAT method
! k_pos      - some integer value 1<=k_pos<=min(nx,nf)
!              the value of xf(k_pos,i) is used to check if the particle has been 
!              evaluated or not. if xf(k_pos,i)>0, the particle has been evaluated, 
!              otherwise it's not! For a general problem, it requires that there 
!              exists one direction whose value is strictly positive otherwise the 
!              user needs to expand the x array to include such direction. For 
!              combustion problems, usually k_pos=nx-2 which corresponds to
!              enthalpy.
! op_frq     > 0, integer, output statistics from x2f_mpi to x2fstats_#.out every
!              op_frq step  
!            < 0, no output
!
! ------------ Other parameters
!    gat     - global number of attempts
!     n      - total number of methods
!  nspec1    - used to specify the size of second dimension of info array
!  nspec2    - used to specify the size of second dimension of rinfo array
!att_max_deft- default maximum number of attempts
!    

implicit none

integer, parameter               :: att_max_deft = 8
integer, intent(in)              :: nx, nf, ldxf, nv, mode, nSR2, istep
integer, intent(inout)           :: info_imp(10)
real(kind(1.d0)), intent(inout)  :: xf(ldxf,nv)

external                         :: x2f_cirxn, x2f_setmethod 
logical, save                    :: init=.false., impinit = .false.
integer, save                    :: dist_strat(3), atmps(3), k_pos, gat, &
                                    n, n_spec1, n_spec2, nolocal, ntmp, op_frq

integer, allocatable             :: info(:,:)
real(kind(1.d0)), allocatable    :: rinfo(:,:), xf_nolocal(:,:)
real(kind(1.d0))                 :: rinfo_xf(1), stats(100)
integer                          :: info_xf(1)
integer                          :: i, nproc, att_max

if ( .not.impinit ) then
   nolocal = info_imp(1)
   att_max = info_imp(2)
   
   impinit = .true.
end if

   ! Initialize pre-set strategies                                     
   if ( .not.init ) then     
      if ( mode==1 ) then
         dist_strat(1:3) = (/0, 0, 0/)
         atmps(1:3)      = (/0, 0, 1/)
      elseif ( mode==2 ) then
         dist_strat(1:3) = (/0, 0, 1/)
         atmps(1:3)      = (/0, 0, 1/)
      elseif ( mode==3 ) then
         call x2f_mpi_nproc( nproc )
         ntmp = nproc-nolocal
         dist_strat(1:3) = (/0, 4, 1/)
         atmps(1:3)      = (/0, ntmp, 1/)
      elseif ( mode==4 ) then         
         if ( att_max<=0 ) att_max = att_max_deft
         call x2f_mpi_nproc( nproc )
         if ( nproc-nolocal>att_max ) then
            ntmp = att_max
         else
            ntmp = nproc-nolocal
         end if
         dist_strat(1:3) = (/0, 4, 1/)
         atmps(1:3)      = (/0, ntmp, 1/)
      elseif ( mode==5 ) then
         call x2f_mpi_nproc( nproc )
         ntmp = nproc-nolocal
         dist_strat(1:3) = (/4, 4, 1/)
         atmps(1:3)      = (/ntmp, min(nSR2,ntmp), 1/)
      elseif ( mode>5.or.mode<1 ) then
         print *, 'mode =', mode, 'has not implemented yet!'
         stop
      end if

      k_pos   = nx - 2 
      k_pos   = 1  
      op_frq  = -1
      gat     = sum( atmps )
      n       = size( atmps, 1 )
      n_spec1 = 10
      n_spec2 = 50   
      init = .true.
   
      !write the mode, distribution strategies and number of attempts to file
      write(*,'(A,10I6)') 'mode=', mode
      write(*,'(A,10I6)') 'nolocal', nolocal
      write(*,'(A,10I6)') 'distribution strategies:', dist_strat
      write(*,'(A,10I6)') 'number of attempts     :', atmps
   end if
   
   !specify info, rinfo, info_xf and rinfo_xf arrays 
   allocate(  info(gat,n_spec1) )
   allocate( rinfo(gat,n_spec2) )

   call construct_ary( dist_strat, atmps, n, gat, op_frq, n_spec1, n_spec2, &
                       nolocal, info, rinfo, info_xf, rinfo_xf )
   
   call x2f_mpi( nx, nf, ldxf, nv, xf, k_pos, info, rinfo, info_xf, rinfo_xf, &
                 gat, n_spec1, n_spec2, x2f_cirxn, x2f_setmethod, stats ) 
            
   deallocate( info  )
   deallocate( rinfo )
   
end subroutine ISAT_MP

   subroutine construct_ary( dist_strat, atmps, n, gat, op_frq, n_spec1, &
                             n_spec2, nolocal, info, rinfo, info_xf, rinfo_xf )
   implicit none
   integer, intent(in)    :: n, dist_strat(n), atmps(n), gat, op_frq, n_spec1, &
                             n_spec2, nolocal
   integer, intent(inout) :: info(gat, n_spec1), info_xf(1)
   real(kind(1.d0)), intent(inout) :: rinfo(gat,n_spec2), rinfo_xf(1)
   integer                :: igat, i, j  

   info     = 0
   rinfo    = 0.d0
   info_xf  = 0
   rinfo_xf = 0.d0   

   ! Sep 16, 2007
   info_xf(1) = atmps(2)
   ! Sep 16, 2007
   igat = 0
   do i = 1, n
      do j = 1, atmps(i)
         igat = igat + 1
		 info(igat,1) = i
		 info(igat,2) = dist_strat(i)
		 info(igat,7) = op_frq
         ! Sep 24, 2007
		 if (j==1) info(igat,6) = 1
		 if (info(igat,2)==4) info(igat,3)=nolocal
	  end do
   end do

   end subroutine construct_ary