module x2f_uran

! Module for x2f_mpi to distribute the vectors uniformly and randomly.
! This guarantees load balancing but has a high degree of message passing.

use MPI
implicit none

contains  !-------------------------------------------------------------

subroutine x2f_uran_init( myrank )

! x2f_uran_init takes myrank as an argument to ensure that each processor
! will generate a different seednum for subsequent calls to random_number

integer, intent(in)  :: myrank
integer, allocatable :: seednum(:)
integer :: seedsize, i

call random_seed( size = seedsize )
allocate( seednum(seedsize) )
seednum = (/ ( 123 * myrank + 7 * i, i = 1, seedsize ) /)
!seednum = (/ ( 123 * myrank + 6 * i, i = 1, seedsize ) /)
call random_seed( put = seednum(1:seedsize) )

return
end subroutine x2f_uran_init

!-----------------------------------------------------------------------

subroutine x2f_uran_p( nproc, mpicomm, myrank, ntogo, nv, p, &
               n_incoming, n_outgoing )

integer, intent(in)    :: nproc, mpicomm, myrank, ntogo, nv
integer, intent(inout) :: p(nv)
integer, intent(out)   :: n_incoming(:), n_outgoing(:)

integer :: n_incoming1(nproc,nproc),ntogo_array(nproc)
integer :: msgsize, total_active, active_pool(nproc), pool_count(nproc)
integer :: j, i, j_pick, p_pick, ierr, n_left
real    :: x1
integer :: iip

! The goal: reassign p(i) with a random number 1:nproc wherever it is initially -1.
! Only reassign p(i) if its initial value is -1; elements equal to zero are unchanged.
! Note, ntogo is the number of elements of p(1:nv) initially equal to -1.
! Elements of p equal 0 where evaluations have already been done via quick try.
! Total assigned to processor j == n_outgoing(j) == ( ntogo / nproc ), for j /= k.

call MPI_Allgather( ntogo,1,MPI_INTEGER,ntogo_array,1,MPI_INTEGER,mpicomm,ierr)
msgsize = ntogo / nproc
n_outgoing(1:nproc) = msgsize
n_left = ntogo - msgsize * nproc
call iuranwor2( msgsize, nproc, myrank, ntogo_array, n_left, n_outgoing )

! Imagine drawing processor assignments one at a time out of a pool of slips of paper.
! At start, exactly n_outgoing(j) slips of paper have the number "j" written on them.

pool_count = n_outgoing

! The count of "j"-labeled slips, pool_count(j), goes down one each time "j" is drawn.
! As the drawing continues, some label "j" will no longer be found in the pool.
! When this happens, pool_count(j) goes to zero; total_active is reduced by one; and 
! active_pool is rearranged so that active_pool(k) /= j for k in 1:total_active.
! We start with total_active = nproc and active_pool(j) = j, in the usual case where
! nv > ntogo > nproc.  But we also want to handle situations where ntogo < nproc...

total_active = 0
do j = 1, nproc
   if( n_outgoing(j) /= 0 ) then
      total_active = total_active + 1
      active_pool(total_active) = j
   end if
end do

do i = 1, nv
   if( p(i) == -1 ) then
      call random_number( x1 )
      ! since 0 <= x1 < 1, j_pick is a bounded integer, 1 <= j_pick <= total_active
      j_pick = int( x1 * total_active ) + 1
      p_pick = active_pool(j_pick)
      pool_count(p_pick) = pool_count(p_pick) - 1
      if( pool_count(p_pick) == 0 ) then
         if( j_pick /= total_active ) then
            ! move an active pool number from the last slot to the newly empty slot
            active_pool(j_pick) = active_pool(total_active)
         end if
         total_active = total_active - 1
      end if
      p(i) = p_pick
   end if
end do

! do a single gather-to-all to determine n_incoming
!call MPI_Allgather( msgsize, 1, MPI_INTEGER, n_incoming, 1, MPI_INTEGER, mpicomm, ierr )
call MPI_Alltoall( n_outgoing, 1, MPI_INTEGER, n_incoming, 1, MPI_INTEGER, mpicomm, ierr )

! n_incoming is very nearly the same for everyone; only n_incoming(k) must be reset
!n_incoming(k) = n_outgoing(k)

return
end subroutine x2f_uran_p

subroutine iuranwor2( msgsize, nproc, myrank, ntogo_array, nleft, n_outgoing )

integer, intent(in)      :: msgsize, nproc, myrank, nleft, ntogo_array(nproc)
integer, intent(inout)   :: n_outgoing(nproc)

integer   :: nsub_tot, nleft_sub, nptemp

if  (myrank==0) then
    n_outgoing(1:nleft) = msgsize + 1
else
    nsub_tot = sum(ntogo_array(1:myrank))
    nptemp   = nsub_tot/nproc
    nleft_sub= nsub_tot - nptemp*nproc
    if  (nproc-nleft_sub>=nleft) then
        n_outgoing(nleft_sub+1:nleft_sub+nleft) = msgsize + 1
    else
        n_outgoing(nleft_sub+1:nproc) = msgsize + 1
        n_outgoing(1:nleft-(nproc-nleft_sub)) = msgsize + 1
    end if
end if

return
end subroutine iuranwor2

end module x2f_uran
